;| --------------------------------------------------------------
;| RetroForth, Release 8
;| --------------------------------------------------------------
init_retroforth:
	call reset				; Init stack
	upsh retroforth_f		        ; Start of bootstrap
	upsh retroforth_f.1 - retroforth_f      ; Length of bootstrap
	call eval 				; Interpret bootstrap 
next 


;| CORE words
;| --------------------------------------------------------------
;| forth   macro
;|  
;| Use:     Switch the active dictionary for compilation
;| Take:    --
;| Return:  --
;| Notes:
; --------------------------------------------------------------

code 'forth', dict_forth
	mov dword [last], flast
next
code 'macro', dict_macro
	mov dword [last], mlast
next


;| --------------------------------------------------------------
;| find   mfind
;| 
;| Use:     Find a word in either the forth (find) or macro
;|          (mfind) dictionaries.
;| Take:	   a #
;| Return:  xt  (if found)
;|          a # (if not found)
;| Notes:   Returns TRUE if found, FALSE if not found.
; --------------------------------------------------------------

code 'mfind', mfind	        ;
	push ebx	        ;
	mov ebx, mlast	        ;
	jmp find.o	        ;
code 'find', find	        ;
	push ebx                ;
	mov ebx, flast          ;
.o:	upop ecx                ;
	push esi
	mov esi,eax             ;
	mov dh,  byte [eax]     ; get first char to do quick compare:
	mov dl, cl              ; and the length; do both compares at once!
.a:	mov ebx,[ebx]           ;
	test ebx,ebx            ;
	jz .end                 ; end of wordlist
	lea edi, [ebx+9]        ; point edi at the string
	cmp dx, [edi-1]
	jne .a
.len:	push esi                ; same length, so do the compare
	push ecx                ;
	repe cmpsb              ;
	pop ecx                 ;
	pop esi                 ;
	jne .a                  ;
	mov eax,[ebx+4]         ; exact match: return XT
	clc                     ;
.ret:   pop esi
	pop ebx                 ;
	ret
.end:	pop esi
	upsh ecx                ; no matches
	stc                     ;
	pop ebx
next


;| --------------------------------------------------------------
;| >number
;|  
;| Use:      Convert a string to an integer on the stack
;| Takes:    a #
;| Returns:  n    (if a valid number in the current base)
;|           a #  (if not a valid number in this base)
;| Notes:    Returns TRUE if the conversion worked; FALSE if it
;|           failed.
; --------------------------------------------------------------

code '>number', number	        ;
	push dword [base]       ;
	mov ecx,eax	        ; n   (keep on stack in case of failure)
	mov ebx,[esi]	        ; a
	dup		        ;
	xor eax,eax	        ; the number
	xor edx,edx	        ; temp
	mov dl,[ebx]	        ;
	cmp dl,45	        ; -     Sign prefix
	pushf		        ;
	jne .a		        ;
	inc ebx 	        ;
	dec ecx 	        ;
	mov dl,[ebx]	        ;
.a:	sub dl,35	        ; $%&'  Base prefix
	cmp dl,5	        ;
	ja .b		        ;
	mov dl,[edx+bases]      ;
	mov byte [base],dl      ;
	inc ebx 	        ;
	dec ecx 	        ;
.b:	mov dl,[ebx]	        ; digits
	inc ebx 	        ;
	call digit	        ;
	jc .err 	        ;
	loop .b 	        ;
	jmp .c		        ;
.err:	add esp, 4	        ;
	drop		        ;
	stc		        ;
	jmp .ret	        ;
.c:	popf		        ;
	jne .d		        ;
	neg eax 	        ;
.d:	add esi,8	        ;
	clc		        ;
.ret:	pop dword [base]        ;
next

digit:	cmp byte [base],255     ;
	je .x10 	        ;
	cmp dl,57	        ; 9
	jbe .a		        ;
	and dl,5Fh	        ; uppercase
	cmp dl,65	        ; throw out chars between '9' and 'A'
	jb .err 	        ;
	sub dl,7	        ;
.a:	sub dl,48	        ; 0
	cmp dl, byte [base]     ;
	jb .x10 	        ;
.err:	stc		        ; not a digit
	ret		        ;
.x10:	imul eax,[base]         ;
	add eax,edx	        ;
	clc		        ;
next


;| --------------------------------------------------------------
;| interpret
;|  
;| Use:     Process input from the TIB
;| Take:    --
;| Return:  --  (results of evaluation)
;| Notes:   'eval' also uses this internally
; --------------------------------------------------------------

code 'interpret', interpret
.o:	call query	        ; ( Get a WORD )
.word:	upsh 32 	        ; ( Push a <SPACE> onto the stack )
	call parse	        ; ( Parse until we find <SPACE> )
	jnz .find	        ; ( Look for the end of the line )
	drop		        ; ( DROP two numbers )
	drop		        ;
	next
.find:	call find	        ; ( See if we can find the word )
	jnc .exec	        ; ( Yes? Interpret it )
	call number	        ; ( No? Then make it a number )
	jnc .word	        ; ( Loop back )
	call [notfound]	; 
	next
.exec:	upop edi	        ; ( This is where we actually call the )
	call edi	        ; ( words we find )
	jmp .word	        ; ( And Loop back )



; --------------------------------------------------------------
; Internal: Gets input, sets TIB, TP, and >IN
; Used by:  interpret, eval
; Notes:    query_mem is never called directly
;           These routines handle CR, LF, TAB, and BACKSPACE
; --------------------------------------------------------------

query:			        ;
	mov dword ecx,[source]  ;
	or ecx,ecx	        ;
	jnz query_mem	        ;
	mov dword [tp],tib      ; Reset TP, TIN
	mov dword [tin],tib     ;
.a:	call key	        ; Get a keypress
	cmp al, 13	        ;
	je .cr		        ;
	cmp al, 10	        ;
	je .cr		        ;
	cmp al, 9		; Was it a TAB (evil!!!)
	je .tab			;
        cmp al, 8		; Perhaps it's a backspace?
        jne .b			;
        mov dword eax,[tp]	;
        sub eax, 1		;
        mov dword [tp],eax	;
        sub edi, 1		;
	drop			;
        jmp .a			;
.b:	xchg edi,[tp]	        ; Store the char in the TIB
	stosb		        ;
	xchg [tp],edi	        ;
	drop		        ;
	jmp .a		        ; And Loop back around
.tab:	mov al, 32		; Convert it to a SPACE
	jmp .b			;
.cr:	drop			; We just drop the CR/LF
	ret			; and exit

query_mem:		        ;
	mov edi,[tin]	        ;
	upsh edi	        ; Input pointer
	dup		        ;
	cmp byte [edi], 9	; TAB
	jne .notab		;
	mov byte [edi], 32	; Make the TAB a SPACE
.notab:	cmp byte [edi],10       ; Skip LF
	jne .a			;
	inc edi 	        ; Increase our pointer
.a:	mov [tin],edi	        ;
	sub ecx,edi	        ; Remaining length
	jbe .eof	        ;
	mov al,10	        ; Line Feed
	repne scasb	        ;
	mov eax,edi	        ;
	jne .b		        ;
	dec eax 	        ;
	cmp Byte [eax-1],13     ; We make CR optional
	jne .b		        ;
	dec eax 	        ;
.b:	upop [tp]	        ;
	drop		        ;
	next		        ;
.eof:	drop		        ;
	drop		        ;
	add esp,4	        ; Discard caller
	pop Dword [tp]	        ;
	pop Dword [tin]         ;
	pop Dword [source]      ;
next


;| --------------------------------------------------------------
;| eval
;|  
;| Use:     Evaluate a string
;| Take:    a #
;| Return:  -- (the results of the evaluation are left on the
;|              stack)
;| Notes:
; --------------------------------------------------------------

code 'eval', eval	        ; This takes an address & count to eval
	push Dword [source]     ; Save "source"
	push Dword [tin]        ; Save ">in"
	push Dword [tp]         ; Save "tp"
	add eax,[esi]	        ;
	upop [source]	        ; New "source"
	upop [tin]	        ; New ">in"
.a:	call query
	call interpret.word
	jmp .a


;| --------------------------------------------------------------
;| 1,   2,   3,   ,
;| 
;| Use:     Inline 1, 2, 3, or 4 bytes to HERE
;| Take:    n
;| Return:  --
;| Notes:
; --------------------------------------------------------------

code ',', comma 	        ; comma (,) saves a value to "here"
	mov ecx,4	        ; By default, it uses a dword (4 bytes)
.a:	mov edx,[h]	        ;
	mov [edx],eax	        ;
	drop		        ;
	add edx,ecx	        ;
	mov [h],edx	        ;
	mov dword [tail], -1	; Disable tail recursion
next
code '1,', comma1	        ; comma1 (1,) saves 1 byte to "here"
	mov ecx,1	        ;
	jmp comma.a
code '2,', comma2	        ; comma2 (2,) saves 2 bytes to "here"
	mov ecx,2	        ;
	jmp comma.a
code '3,', comma3	        ; comma3 (3,) saves 3 bytes to "here"
	mov ecx,3	        ;
	jmp comma.a


; --------------------------------------------------------------
; Internal: Handle a numeric literal in compiled code
; Used by:  literal
; Notes:    This isn't the most efficient solution, but it's not
;           intended to be. It was targetted to design
;           flexibility.
; --------------------------------------------------------------
dolit:	dup		        ; This is where we handle literals.
	mov eax,[esp]	        ;
	mov eax,[eax]	        ;
	add dword [esp],4       ;
next


;| --------------------------------------------------------------
;| entry
;| 
;| Use:     Create a new dictionary entry from a string
;| Take:    a #
;| Return:  --
;| Notes:   This does not define a body for the entry, it just
;|          creates an entry pointed to HERE
; --------------------------------------------------------------

code 'entry', _entry
	push ecx	        ;
	mov edi,[d]	        ; LFA
	mov ecx,[last]	        ; last
	mov edx,[ecx]	        ;
	mov [edi],edx	        ; LFA= [last]
	mov [ecx],edi	        ; last= LFA
	mov ecx,[h]	        ;
	mov [edi+4],ecx         ; CFA= here
	mov [edi+8],al	        ; Length
	add edi,9	        ;
	upop ecx	        ;
	push esi	        ;
	mov esi,eax	        ;
	rep movsb	        ;
	mov [d],edi	        ; d= d+9+length
	pop esi 	        ;
	pop ecx 	        ;
	drop
next


;| --------------------------------------------------------------
;| ]
;|  
;| Use:     The compiler
;| Take:    --
;| Return:  --
;| Notes:   This is where the compilation of a word actually
;|          takes place
; --------------------------------------------------------------

code ']', rbracket ; This is the actual compiler loop
.word:	upsh 32 	        ; Space
	call parse	        ; Scan for a word name or number
	jnz .find	        ; n=0 means EOL
	drop		        ; 2drop
	drop		        ;
	call query	        ;
	jmp .word	        ; Was a 'ret'
.find:	call mfind	        ;
	jnc .exec	        ; Macros
	call find	        ;
	jnc .com	        ; Regular words
	call number	        ;
	jnc .lit	        ; Numbers
	call [notfound]		; Go figure..
	jmp  undo_compile	; And get rid of the bad definition
.exec:	upop edi	        ; Execute words in 'macro'
	call edi	        ;
	jmp .word	        ;
.com:	call compile	        ; Compile a call to words in 'forth'
	jmp .word	        ;
.lit:	call literal		; Literal
	jmp .word	        ;


; --------------------------------------------------------------
; Internal: Undo the last compilation
; Used by:  ]
; Notes:    Resets HERE, and LAST to their prior state
; --------------------------------------------------------------

undo_compile:			;
	mov ebx, dword [old_h]	;
	mov dword [h], ebx	;
	mov ebx, dword [old_last]
	mov ebx, [ebx]
	mov dword [last], ebx	;
	add esp, 4		;
next


;| --------------------------------------------------------------
;| compile
;|  
;| Use:     Compile a CALL to a word
;| Take:    a
;| Return:  --
;| Notes:   The CALL is relative, not absolute
; --------------------------------------------------------------

code 'compile', compile         ; This routine compiles in CALL's to
	sub eax,[h]	        ; words. Address is passed in EAX
	sub eax,5	        ;
	upsh 0xE8	        ; CALL opcode
	call comma1	        ; 0xE8
	call comma	        ; ADDRESS
	mov dword [tail], 0	; Enable tail-call
next


;| --------------------------------------------------------------
;| [
;|  
;| Use:     Stop a compile, drop back to the interpreter
;| Take:    --
;| Return:  --
;| Notes:   Does not terminate the definition, use ; for that.
; --------------------------------------------------------------

mcode '[', lbracket	        ; Switch back to the interpreter
	add esp,4	        ;
next



;| --------------------------------------------------------------
;| ;;
;| 
;| Use:     compile an exit to a definition
;| Take:    --
;| Return:  --
;| Notes:   If the last thing compiled was a call, change to a
;|          jmp. Otherwise just compiles in a ret.
; --------------------------------------------------------------

mcode ';;', ssemi	        ; Exit a word (; will call this!)
	mov edx,[h]	        ;
	sub edx,5	        ;
	cmp byte [edx],0xE8     ; Was the last thing compiled a CALL?
	jnz .a		        ; No, skip the next line
	cmp dword [tail],0	; See if we should compile a tail-call
	jnz .a			; No? skip it and compile a ret
	inc byte [edx]	        ; Yes, change to JMP
	next		        ; and exit
.a:	mov byte [edx+5],0xC3   ; If not a CALL, compile in a RET
	inc dword [h]	        ;
next


;| --------------------------------------------------------------
;| ;
;|  
;| Use:     End a definition
;| Take:    --
;| Return:  --
;| Notes:   Calls ;; and then jumps to [
; --------------------------------------------------------------

mcode ';', semi ; Compile in an exit to the current word
	call ssemi	        ; And go back to the interpreter
	jmp lbracket	        ;


;| --------------------------------------------------------------
;| :
;| 
;| Use:     Main compiler wrapper
;| Take:    --
;| Return:  --
;| Notes:   Creates an entry (saving the old state of HERE and
;|          LAST), then jumps to ]
; --------------------------------------------------------------

code ':', colon 	        ; Ok, this is the entry to the compiler
	mov ebx, dword [last]	;
	mov dword [old_last], ebx
	mov ebx, dword [h]	;
	mov dword [old_h], ebx	;
	upsh 32			;
	call parse		;
	call _entry	        ; * Create a new word
	jmp rbracket	        ; * And jump to the real compiler


;| --------------------------------------------------------------
;| literal
;|  
;| Use:     Compiles a literal into a definition
;| Take:    n
;| Return:  --
;| Notes:   
; --------------------------------------------------------------

mcode 'literal', literal        ; Compile in a literal
	upsh dolit	        ;
	call compile	        ;
	call comma	        ;
next


;| --------------------------------------------------------------
;| #
;| 
;| Use:     Display the unsigned value of the TOS
;| Take:    n
;| Return:  --
;| Notes:   . and u. are written around this. # does not leave a
;|          trailing space.
; --------------------------------------------------------------

code '#', print 	        ;
	push edx	        ;
	push edi	        ;
	mov edi,esi	        ; edi = buffer (in stack space)
	sub edi,4	        ;
.a:	xor edx,edx	        ;
	div dword [base]        ;
	add dl,'0'	        ;
	cmp dl,'9'	        ;
	jbe .b		        ;
	add dl,7+32	        ;
.b:	dec edi 	        ;
	mov [edi],dl	        ;
	or eax,eax	        ;
	jnz .a		        ;
	mov eax,edi	        ; Print
	upsh esi	        ;
	sub eax,edi	        ; # of digits
	call type	        ;
	pop edi 	        ;
	pop edx 	        ;
next


;| --------------------------------------------------------------
;| parse
;|  
;| Use:     Parse input stream
;| Take:    <char>
;| Return:  a #
;| Notes:   Parses TIB until it encounters <char> or the end of a
;|          line.
; --------------------------------------------------------------

code 'parse', parse	        ;
	mov edi,[tin]	        ; Pointer into TIB
	mov ecx,[tp]	        ;
	sub ecx,edi	        ;
	inc ecx 	        ;
	repe scasb	        ;
	dec edi 	        ;
	inc ecx 	        ;
	dup		        ;
	mov [esi],edi	        ; a
	repne scasb	        ;
	mov eax,edi	        ;
	dec eax 	        ;
	sub eax,[esi]	        ; n
	mov [tin],edi	        ;
next


;| --------------------------------------------------------------
;| reset
;|  
;| Use:     Reset the stack to the default starting point
;| Take:    --
;| Return:  --
;| Notes:   Also replaces all values on the stack with 0's
; --------------------------------------------------------------

code 'reset', reset	        ; This word is used to reset the stack
	mov ecx, 2048		;
	mov esi, s0		;
	sub esi, 1024		;
.empty:	mov byte [esi], 0	;
	inc esi			;
	loop .empty		;
	mov esi, s0		;
	xor eax, eax		;
next


;| --------------------------------------------------------------
;| create
;| 
;| Use:     create a variable
;| Take:    --
;| Return:  --
;| Notes:   parses the input stream for the name
; --------------------------------------------------------------

code 'create', create
	upsh 32
	call parse
	call _entry
	upsh dovar
	jmp compile

dovar:	dup
 	pop eax
next


;| --------------------------------------------------------------
;| does>
;|  
;| Use:     Modify the runtime behavor of a 'create'd word
;| Take:    --
;| Return:  addr
;| Notes:   changes 'call dovar' to 'call <code after does>'
;|          Adds some code to push the address of the data part
;|          to the stack. 
; --------------------------------------------------------------

mcode 'does>', does             ;
        upsh pdoes              ;
        call compile            ;
        upsh dodoes             ;
        jmp compile             ;

dodoes: dup                     ; Handle DOES>
        pop eax                 ;
        xchg eax,[esp]          ;
next

pdoes:  dup                     ;
        mov eax,[last]          ; eax= header
        mov eax,[eax]           ;
        mov eax,[eax+4]         ; eax= CFA
        pop ebx                 ;
        sub ebx,eax             ;
        sub ebx,5               ;
        mov [eax+1],ebx         ; Change "call dovar" to "call <code after does>"
        drop                    ;
next


;| --------------------------------------------------------------
;| last
;|  
;| Use:     Returns the address of the last dictionary entry
;| Take:    --
;| Return:  addr
;| Notes:
; --------------------------------------------------------------

code 'last', var_last
	upsh [last]
next


;| --------------------------------------------------------------
;| there
;|  
;| Use:     Returns the address of the block buffer
;| Take:    --
;| Return:  addr
;| Notes:
; --------------------------------------------------------------

code 'there', there
	upsh blk
next


;| --------------------------------------------------------------
;| tib
;|  
;| Use:     Push the address of the TIB
;| Take:    --
;| Return:  addr
;| Notes:
; --------------------------------------------------------------

code 'tib', _tib
	upsh tib
next

; --------------------------------------------------------------
var 'h0', h, h0		        ; h0     (pointer to HERE)
var 'base', base, 10	        ; base   current numeric base
var '>in', tin, 0	        ; >IN    Pointer into the TIB
var 'word?', notfound, 0	; ?word  Handler for errors
old_h		dd 0		;
old_last	dd 0		;
d		dd d0		;
tail		dd 0		; Allow tail-calls?
source		dd 0	        ; Evaluate from RAM or KBD
tp		dd tib	        ; TP (pointer to input buffer)
bases		db 10,16,2,8,255; #dec $hex %bin &oct 'ascii
; --------------------------------------------------------------
include 'core.dict'
last  dd flast			; Last word in dictionary
flast dd forth_link	        ; Last word in 'forth'
mlast dd macro_link	        ; Last word in 'macro'
; --------------------------------------------------------------
retroforth_f: file "core.f"     ; Include the bootstrap code
retroforth_f.1: 		;
; --------------------------------------------------------------
section "bss"
; --------------------------------------------------------------
    rb 1024			; 256 cells  \ Total of 512 cells
s0  rb 1024			; 256 cells  / 2,048 bytes
tib rb 1024			; Text Input Buffer (1 KiB)
d0  rb 40000h			; Dictionary     (256 KiB)
h0  rb 100000h                  ; Code ("heap")  (1 MiB)
blk rb 128*1024			; Block buffer   (128 KiB)
; --------------------------------------------------------------
